
module FFJ.AuxFunc where  {
	import FFJ.Syntax;
	import FFJ.Program;
	import Maybe;
	import Debug.Trace;
	
	fields :: CT -> ExtType -> [Field];
	fields ct ce = case (classdef ct ce) of {
		Just (False,"Object","",[],[]) -> [];
		Just (True,c,d,fs,ms) -> case (getPred ct ce) of {
			Just (e,omega) -> (fields ct (e,omega)) ++ fs;
			Nothing -> fs;
		};
		Just (False,c,d,fs,ms) -> (fields ct (getLast ct d)) ++ fs;
		Nothing -> [];
	};
	
	getFieldType :: Identifier -> [Field] -> Type;
	getFieldType f ((c,g):gs) | (f == g) = c
	                          | otherwise = (getFieldType f gs);
	getFieldType f [] = error ("Field " ++ f ++ " does not exist! (getFieldType)");
	
	mbody :: CT -> Identifier -> ExtType -> Maybe MethodBody;
	mbody ct m ce = case (classdef ct ce) of {
		Just (False,c,d,fs,ms) -> case mdef m ms of {
			Just (o,r,n,ps,b) -> Just (snd (unzip ps),b);
			Nothing -> if (length d) /= 0 then mbody ct m (getLast ct d) else error ("Method " ++ m ++ " not found in class " ++ c ++ "! (mbody)");
		};
		Just (True,c,d,fs,ms) -> case mdef m ms of {
			Just (o,r,n,ps,b) -> Just (snd (unzip ps),b);
			Nothing -> case (getPred ct ce) of {
				Just (e,omega) -> mbody ct m (e,omega);
				Nothing -> error ("Method " ++ m ++ " not found in class " ++ c ++ "! (mbody)");
			};
		};
		Nothing -> error ("Class " ++ (show ce) ++ " not found in class table! (mbody)");
	};

	mdef :: Identifier -> [Method] -> Maybe Method;
	mdef n ((o,r,m,ps,b):xs) | (m == n) = Just (o,r,m,ps,b) | otherwise = mdef n xs;
	mdef n [] = Nothing;

	
	mtype :: CT -> Identifier -> ExtType -> Maybe Signature;
	mtype ct m ce = case (classdef ct ce) of {
		Just (False,c,d,fs,ms) -> case mdef m ms of {
			Just (False,r,n,ps,b) -> Just (fst (unzip ps),r);
			Just (True,r,n,ps,b) -> mtype ct m (getLast ct d);
			Nothing -> if c /= "Object" then mtype ct m (getLast ct d) else Nothing;
		};
		Just (True,c,d,fs,ms) -> case mdef m ms of {
			Just (False,r,n,ps,b) -> Just (fst (unzip ps),r);
			Just (True,r,n,ps,b) -> mtype ct m (let pre = (getPred ct ce) in 
			                            if (isJust pre) 
			                            then fromJust pre 
			                            else error ("No predecessor of " ++ (show ce) ++ " found in class table! (mtype)" ));
			Nothing -> mtype ct m (let pre = (getPred ct ce) in 
			                            if (isJust pre) 
			                            then fromJust pre 
			                            else error ("No predecessor of " ++ (show ce) ++ " found in class table! (mtype)" ));
		};
		Nothing -> error ("Class " ++ (show ce) ++ " not found in class table! (mtype)");
	};
	
	override :: CT -> Identifier -> ExtType -> Signature -> Bool;
	override ct m c (ps, r) | (isJust (mtype ct m c)) = (ps == (fst (fromJust (mtype ct m c))) && r == (snd (fromJust (mtype ct m c)))) | otherwise = False;

	introduce :: CT -> Identifier -> ExtType -> Bool;
	introduce ct m c = case (mtype ct m c) of {
		Just sgn -> False;
		Nothing -> True;
	};

	
--	getTypeFromClass :: CT -> Class -> Type;
--	getTypeFromClass ct (_,c,_,_,_) = c;
	
}