
module FFJ.Typing where {
	import FFJ.Syntax;
	import FFJ.Program;
	import FFJ.AuxFunc;
	import List;
	import Maybe;
	import Debug.Trace;
	
	type Gamma = [(Variable,Type)];

	allOK :: CT -> PT -> Bool;
	allOK ct pt = (ctOK ct) && (classesOK ct ct);
	-- && (programsOK ct pt);
	
	ctOK :: CT -> Bool;
	ctOK ct = (objectIsOK ct) && (objectIsUnique ct) && (hasNoCycles ct ct);

	objectIsOK :: CT -> Bool;
	objectIsOK ct = case (lookup ("Object","None") ct) of {
		Just (o,c,d,f,m) -> ((length f == 0) && (length m == 0) && (d == ""));
		Nothing -> False; 
	};

	objectIsUnique :: CT -> Bool;
	objectIsUnique ct = (length (findIndices (== ("Object","None")) (fst (unzip ct))) == 1);

	hasNoCycles :: CT -> CT -> Bool;
	hasNoCycles ct ((c,cd):xs) = case cd of {
		(False,_,_,_,_) -> (not (inducesCycle ct c (getFirst ct (superclass cd)))) && hasNoCycles ct xs;
		(True,_,_,_,_) -> hasNoCycles ct xs;
	};
	hasNoCycles ct [] = True;
	
	inducesCycle :: CT -> ExtType -> ExtType -> Bool;
	inducesCycle ct c d | ((c == ("Object","None")) || (d == ("Object","None"))) = False
                        | (c == d) = True 
	                    | otherwise = if (isJust cd) then inducesCycle ct c sc else error ("Class " ++ (show c) ++ " not found in class table! (inducesCycle)")
	                    where { cd = (classdef ct d); sc = (getFirst ct (superclass (fromJust cd))) };


	programsOK :: CT -> PT -> Bool;
	programsOK ct (p:ps) = (programOK ct p) && (programsOK ct ps);
	programsOK ct [] = True;

	programOK :: CT -> (PTEntry, Term) -> Bool;
	programOK ct (_,t) = isJust (classdef ct (getFirst ct (termType ct [] t)));

	classesOK :: CT -> CT -> Bool;
	classesOK ct (((n,feat),c):cs) = (classOK ct c feat) && (classesOK ct cs);
	classesOK ct [] = True;
	
	classOK :: CT -> Class -> Feature -> Bool;
	classOK ct (False,c,d,fs,ms) feat = case (getPred ct (c,feat)) of {
		Just pre -> False;
		Nothing -> (methodsOK ct ms (c,feat) d);
	};
	
	
	classOK ct (True,c,d,fs,ms) feat = case (getPred ct (c,feat)) of {
		Just pre -> (methodsOK ct ms (c,feat) d);
		Nothing -> False;
	};
	
	equalLists :: [String] -> [String] -> Bool;
	equalLists (a:as) (b:bs) = (a == b) && (equalLists as bs);
	equalLists [] [] = True;
	equalLists [] _ = False;
	equalLists _ [] = False;
	
	methodsOK :: CT -> [Method] -> ExtType -> Type -> Bool;
	methodsOK ct (m:ms) c d = if (d /= "") then (methodOK ct m c d) && (methodsOK ct ms c d) else (methodsOK ct ms c d);
	methodsOK _ [] _ _ = True;
	
	methodOK :: CT -> Method -> ExtType -> Type -> Bool;
	methodOK ct (False,r,m,ps,t) (c,phi) d = case (classdef ct (c,phi)) of {
		Just (False,_,_,_,_) -> (introduce ct m (getLast ct d)) && (subtype ct (getFirst ct (termType ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t)) (getFirst ct r));
		Just (True,_,_,_,_) -> let pre = (getPred ct (c,phi)) in if (isJust pre) then (introduce ct m (fromJust pre)) && (subtype ct (getFirst ct (termType ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t)) (getFirst ct r)) else error ("No predecessor of " ++ (show (c,phi)) ++ " found in class table!" );
		Nothing -> error ("Type error: class declaration of " ++ c ++ " does not exist!");
	};
	methodOK ct (True,r,m,ps,t) (c,phi) d = case (classdef ct (c,phi)) of {
		Just (False,_,_,_,_) -> (override ct m (getLast ct d) (fst (unzip ps),r)) && (subtype ct (getFirst ct (termType ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t)) (getFirst ct r));
		Just (True,_,_,_,_) -> let pre = (getPred ct (c,phi)) in if (isJust pre) then (override ct m (fromJust pre) (fst (unzip ps),r)) && (subtype ct (getFirst ct (termType ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t)) (getFirst ct r)) else error ("No predecessor of " ++ (show (c,phi)) ++ " found in class table!" );
		Nothing -> error ("Type error: class declaration of " ++ c ++ " does not exist!");
	};

--		
--		

	
	termType :: CT -> Gamma -> Term -> Type;
	termType ct gamma (Var v) = case (lookup v gamma) of {
		Just c -> c;
		Nothing -> error ("Type error: type of variable " ++ v ++ " not known!");
	};
	
	termType ct gamma (Proj t f) = getFieldType f (fields ct (getLast ct (termType ct gamma t)));
	
	termType ct gamma (Inv t m ts) = let c0 = getLast ct (termType ct gamma t) in case (mtype ct m c0) of {
		Just (ds,c) -> if (subtypesT ct (map (termType ct gamma) ts) ds) then c else error ("Type error: signature of method " ++ m ++ " does not match!");
		Nothing -> error ("Type error: method " ++ m ++ " does not exist in class " ++ (show c0) ++ "!");
	};
	
	termType ct gamma (Inst c ts) = if (subtypesT ct (map (termType ct gamma) ts) (fst (unzip (fields ct (getLast ct c))))) then c else error ("Type error: arguments do not match the fields of class " ++ c ++"!");
	
	termType ct gamma (Cast c t) | (subtype ct (getFirst ct d) (getFirst ct c)) = c
	                             | (subtype ct (getFirst ct c) (getFirst ct d)) && (c /= d) = c
	                             | otherwise = error ("Type error: stupid cast of type " ++ d ++ " to type " ++ c ++ "!")
	                             where { d = (termType ct gamma t); };
	
	--wellformed :: Method -> Bool;
	

}
