module FFJPL.Wellformedness where {
	import FFJ.Syntax;
	import FFJPL.Definitions;
	import FFJPL.FeatureModel;
	import FFJPL.Predicates;
	import FFJPL.Navigation;
	import FFJPL.Functions;
	import FFJPL.Subtype;
	import FFJPL.TermTyping;	
	import Debug.Trace;
	
	classesAndRefinementsOK :: FeatureModel -> CT -> Bool;
	classesAndRefinementsOK fm ct = classesAndRefinementsOKIter fm ct ct;
	
	classesAndRefinementsOKIter :: FeatureModel -> CT -> CT -> Bool;
	classesAndRefinementsOKIter fm ct (((c,phi),decl):cs) = if (classOrRefinementOK fm ct decl phi) then (classesAndRefinementsOKIter fm ct cs) else (error ("Type error: Class or refinement (" ++ c ++ "," ++ phi ++ ") is not well-formed!"));
	classesAndRefinementsOKIter fm ct [] = True;
	
	classOrRefinementOK :: FeatureModel -> CT -> Class -> Feature -> Bool;
	classOrRefinementOK fm ct (False,c,d,fs,ms) phi = (c == "Object" && phi == "Base") ||
	                                                  (if (validRef fm ct [phi] d) then True else error ("Type error: Supertype " ++ d ++ " of qualified type " ++ phi ++ "." ++ c ++ " is not always present!")) &&
	                                                  (if (all (\t -> validRef fm ct [phi] t) (fst (unzip fs))) then True else error ("Type error: Not all field types of qualified type "  ++ phi ++ "." ++ c ++  " are valid!   (" ++ (show fs) ++ ")")) &&
	                                                  (if (all (\f -> introduceField fm ct [phi] f (getLast ct d)) (snd (unzip fs))) then True else error ("Type error: Some field of qualified type " ++ phi ++ "." ++ c ++ " is already present in its supertype!   (" ++ (show fs) ++ ")")) &&
	                                                  (if (introduceClass fm ct [phi] (c,phi)) then True else error ("Type error: Class " ++ c ++ " is introduced multiple times!   (" ++ (show (getIntroductions fm ct c)) ++ ")")) &&
	                                                  (all (\mdef -> methodOK fm ct mdef (c,phi)) (ms));

	classOrRefinementOK fm ct (True,c,_,fs,ms) phi = (if (all (\t -> validRef fm ct [phi] t) (fst (unzip fs))) then True else error ("Type error: Not all field types of qualified type "  ++ phi ++ "." ++ c ++  " are valid!   (" ++ (show fs) ++ ")")) &&
	                                                 (if (all (\f -> introduceField fm ct [phi] f (getPred ct (c,phi))) (snd (unzip fs))) then True else error ("Type error: Some field of qualified type " ++ phi ++ "." ++ c ++ " is already present in its refinement chain!   (" ++ (show fs) ++ ")")) &&
	                                                 (if (refine fm ct [phi] (c,phi)) then True else error ("Type error: Class " ++ c ++ ", refined in feature " ++ phi ++ ", has not been introduced before in the refinement chain or, if so, it is not always present!")) &&
	                                                 (all (\mdef -> methodOK fm ct mdef (c,phi)) (ms));

	                                                  
	methodOK :: FeatureModel -> CT -> Method -> ExtType -> Bool;
	methodOK fm ct (False,r,m,ps,t) (c,phi) = case getClass ct (c,phi) of {
	                                            (False,c,d,fs,ms) -> (if (and (map (\e -> isSubType fm ct e r [phi]) (getTermType fm ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t phi))) then True else error ("Type error: Not all method body types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are always subtypes of the declared result types!")) &&
	                                                                 (if (all (\b -> validRef fm ct [phi] b) (fst (unzip ps))) then True else error ("Type error: Not all parameter types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are valid!")) &&
	                                                                 (if (introduceMethod fm ct [phi] m (getLast ct d)) then True else error ("Type error: Method " ++ m ++ " of qualified type " ++ phi ++ "." ++ c ++ " is already present in the supertype!"));
	                                            (True,c,_,fs,ms) -> (if (and (map (\e -> isSubType fm ct e r [phi]) (getTermType fm ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t phi)))  then True else error ("Type error: Not all method body types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are always subtypes of the declared result types!")) &&
	                                                                (if (all (\b -> validRef fm ct [phi] b) (fst (unzip ps)))  then True else error ("Type error: Not all parameter types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are valid!")) &&
	                                                                (if (introduceMethod fm ct [phi] m (getPred ct (c,phi))) then True else error ("Type error: Method " ++ m ++ " of qualified type " ++ phi ++ "." ++ c ++ " is already present in the refinement chain!"));
	                                          };
	methodOK fm ct (True,r,m,ps,t) (c,phi) = case getClass ct (c,phi) of {
	                                            (False,c,d,fs,ms) -> (if (and (map (\e -> isSubType fm ct e r [phi]) (getTermType fm ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t phi)))  then True else error ("Type error: Not all method body types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are always subtypes of the declared result types!")) &&
	                                                                 (if (all (\b -> validRef fm ct [phi] b) (fst (unzip ps)))  then True else error ("Type error: Not all parameter types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are valid!")) &&
	                                                                 (if (override fm ct [phi] m (getLast ct d) (fst (unzip ps),r)) then True else error ("Type error: For method refinement " ++ m ++ " of qualified type " ++ phi ++ "." ++ c ++ " there is no proper method introduction in the supertype!"));
	                                            (True,c,_,fs,ms) -> (if (and (map (\e -> isSubType fm ct e r [phi]) (getTermType fm ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t phi)))  then True else error ("Type error: Not all method body types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are always subtypes of the declared result types!")) &&
	                                                                (if (all (\b -> validRef fm ct [phi] b) (fst (unzip ps)))  then True else error ("Type error: Not all parameter types of method " ++ m ++ " in " ++ phi ++ "." ++ c ++ " are valid!")) &&
	                                                                (if (override fm ct [phi] m (getPred ct (c,phi)) (fst (unzip ps),r))  then True else error ("Type error: For method refinement " ++ m ++ " of qualified type " ++ phi ++ "." ++ c ++ " there is no proper method introduction in the refinement chain!"));
	                                          };
	
}